/*
    Fichier suphtm.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence.fr

    Traductions des droits d'utilisation dans les fichiers :
    licence.de , licence.en , licence.es , licence.it
    licence.nl , licence.pt , licence.eo , licence.eo-utf


    Ce programme dtruit les mails crits en HTML pur.

    Ce format de mail trs rarement utilis par des particuliers
    est par contre beaucoup plus frquent dans les spams.

    Un message est gnr  l'intention de l'expditeur du mail
    non rcupr.

    Un fichier de configuration est utilis pour se connecter  la
    boite aux lettres et pour slectionner le serveur SMTP servant
     l'envoi des messages.
*/

#define appli   // pour la dclaration de variables globales  l'application

#include <stdio.h>
#include <string.h>
#include "messages.h"
#include "buflect.h"
#include "ficonf.h"
#include "pop.h"
#include "smtp.h"
#include "trtentete.h"
#include "datecour.h"
#include "mailrep.h"


/* prototypes */
void filtremail (int numes);
int  avertir ();
void ajout_adr (char *dest, char *source);
void transf_ligne (char *message);


/* variables globales au source
   (pour viter des tonnes de passages de paramtres) */

char adr_mail [80]; // adresse mail de l'utilisateur
char pref_adr [80]; // prfixe vraie adresse mail pour la masquer aux
                    // robots qui traitent les rponses aux spams
FILE *ftrace;       // Fichier contenant les adresses d'expditeurs refuss
char datecour [10]; // Date courante (pour le fichier d'expditeurs refuss)
int  opts;          // Option -s (mails refuss non signals aux expditeurs)

/* buffers rappelant les caractristiques du message */
char bufDate [80], bufFrom [80], bufTo [80], bufCc [80], bufReply [80],
     bufSubject [80], bufContent [80], bufRetPath [80];

/* compteurs */
int  conserves = 0;
int  supprimes = 0;
int  avertis = 0;
int  non_avertis = 0;



/* programme principal */

int main (int nbarg, char *varg[])
{
    char serv_smtp [80]; // nom du serveur smtp utilis
    FILE *fconf;         // descripteur du fichier de configuration
    int  numes, nbmes;   // numro du mail courant et nombre de mails


    // rcupration du nom de l'excutable
    memcom (*varg);

    // se positionner sur le premier argument de suphtm
    nbarg --;
    varg ++;

    // initialisation des options de suphtm
    *pref_adr = '\0';
    ftrace = NULL;
    opts = 0;

    // prise en compte des options de suphtm
    while (nbarg > 1 && **varg == '-')
    {
        // option -p (prfixe dans l'adresse d'expditeur)
        if (varg [0][1] == 'p')
        {
            // prfixe par dfaut
            if (varg [0][2] == 'd')
            {
                strcpy (pref_adr, "suphtm");

                // sauter l'argument trait
                nbarg --;
                varg ++;
            }
            // prfixe pass en paramtre
            else if (varg [0][2] == '\0')
            {
                strcpy (pref_adr, varg [1]);

                // sauter les arguments traits
                nbarg -= 2;
                varg += 2;
            }
        }
        // option -s (pas de mail d'avertissement lorsque mail HTML dtruit)
        else if (strcmp (*varg, "-s") == 0)
        {
            opts = 1;

            // sauter l'argument trait
            nbarg --;
            varg ++;
        }

        // option -t (utilisation d'un fichier trace)
        else if (strcmp (*varg, "-t") == 0)
        {
            // ouvrir le fichier trace en criture fin de fichier
            ftrace = fopen (varg [1], "a");

            // si nom de fichier correct
            if (ftrace)
                // initialiser la date courante
                initdatecour (datecour);
            else
                // sinon, avertir l'utilisateur
                // "Impossible d'crire dans le fichier %s"
                aff_err_arg ("IMPOS_ECR_FICH", varg [1]);

            // sauter les arguments traits
            nbarg -= 2;
            varg += 2;
        }
        // option incorrecte
        else
        {
            // "Option %s incorrecte"
            aff_err_arg ("ERR_OPTION", *varg);
            nbarg = 0;  // on terminera en affichant la syntaxe de la commande
        }
    }

    // aucun traitement si option -p et -s utilises simultanment
    if (*pref_adr && opts)
        // "Option -p et -s incompatibles"
        affiche_err ("COMPAT_OPT_PS");

    // sinon, controle du nombre d'arguments restants
    else if (nbarg == 1)
    {
        // ouvrir le fichier de configuration
        fconf = ouvre_ficonf (*varg);

        if (fconf)
        {
            // connexion sur le compte mail du serveur pop
            if (connect_pop (fconf))
            {
                // rcupration du nombre de mails
                nbmes = nbmails ();

                // on saute la ligne contenant le rpertoire des mails
                while (fgetc (fconf) != '\n')
                    ;

                // recuprer adresse mail utilisateur
                fgets (buf_lect, sizeof (buf_lect), fconf);
                buf_lect [strlen (buf_lect) - 1] = '\0';

                // et la mmoriser
                *adr_mail = '\0';
                ajout_adr (adr_mail, buf_lect);

                // recuprer le nom du serveur smtp
                fgets (serv_smtp, 80, fconf);
                serv_smtp [strlen (serv_smtp) - 1] = '\0';

                // s'il y a des mails  tester
                if (nbmes > 0)
                {
                    // connexion au serveur smtp pour l'envoi des avertissements
                    if (opts || connect_smtp (serv_smtp))
                    {
                        // filtrage de la liste des mails
                        for (numes = 1; numes <= nbmes; numes++)
                            filtremail (numes);

                        // dition d'un rcapitulatif
                        // "\n%d messages conservs, %d supprims\n"
                        printf (message ("BILAN_FILTRAGE"), conserves,
                                                            supprimes);

                        if (supprimes && (opts == 0))
                            // "%d expditeurs avertis, %d non avertis\n"
                            printf (message ("BILAN_AVERTIS"), avertis,
                                                               non_avertis);

                        // fermeture de la connexion smtp si elle tait ouverte
                        if (opts == 0)
                            deconnect_smtp ();
                    }
                }

                // se dconnecter proprement du serveur pop
                deconnect_pop ();

                // si des mails ont t supprims
                if (supprimes)
                    // attendre pour enregistrement correct des suppressions
                    // (vite problme si autre filtre appel juste aprs)
                    sleep (2);
            }

            // on n'a plus besoin du fichier de configuration
            fclose (fconf);
        }
    }
    else
        // "Syntaxe : %s [-pd|-p prfixe|-s]
        //               [-t fichier_adresses] fichier_configuration"
        psyntaxe ("SYNT_SUPHTM");

    // pour faire plaisir  gcc qui veut une fonction main de type int
    return (0);
}


/* vrifie si le message choisi est en HTML pur, et
   le dtruit dans ce cas en avertissant l'expditeur */

void filtremail (int numes)
{
    char bufw [80];  // buffer d'envoi d'une requte
    int  posbuf;     // simple compteur


    // initialisation
    bufDate [0]    = '\0';
    bufFrom [0]    = '\0';
    bufTo [0]      = '\0';
    bufCc [0]      = '\0';
    bufReply [0]   = '\0';
    bufSubject [0] = '\0';
    bufContent [0] = '\0';
    bufRetPath [0] = '\0';

    // message de suivi de droulement
    // "\rAnalyse du mail n %d       "
    printf (message ("ANALYSE_MAIL"), numes);
    fflush (stdout);

    /* demande de lecture de l'entte du message
       On demande 5 lignes du corps du message pour trouver le
       2me Content-type si le premier est multipart/mixed */
    sprintf (bufw, "TOP %d 5", numes);
    env_pop (bufw);

    // lecture et mmorisation des caractristiques du message
    do
    {
        // lire une ligne de l'entte du message
        lire_pop ();

        // convertir les caractres spciaux
        majlignentete ();

        // mmorisation des caractristiques importantes
        if (start ("Date"))
            membuf (bufDate);
        else if (start ("From"))
            membuf (bufFrom);
        else if (start ("To"))
            membuf (bufTo);
        else if (start ("Cc"))
            membuf (bufCc);
        else if (start ("Reply-To"))
            membuf (bufReply);
        else if (start ("Subject"))
            membuf (bufSubject);
        else if (start ("Content-Type"))
            membuf (bufContent);
        else if (start ("Return-Path"))
            membuf (bufRetPath);
    }
    // lecture termine si ligne limite  un .
    while (buf_lect [0] != '.' || buf_lect [1] != '\0');

    // si le mail contenait une ligne Content-Type
    if (*bufContent)
    {
        // positionnement sur le 2me attribut de Content_type
        // ncessaire pour bonne dtection s'il y  0 ou plus
        // d'un blanc aprs le : du Content-Type
        posbuf = 17;

        while (bufContent [posbuf] != '/')
            posbuf++;
    }

    // si le mail est en HTML pur
    if (*bufContent && tolower (bufContent [posbuf+1]) == 'h'
                    && tolower (bufContent [posbuf+2]) == 't')
    {
        // demande de destruction du message
        sprintf (bufw, "DELE %d", numes);
        env_pop (bufw);
        lire_pop ();
        supprimes ++;

        // si utilisation d'un fichier trace
        if (ftrace)
            // mmoriser l'adresse de l'expditeur dans ce fichier
            fprintf (ftrace, "%s%s\n", datecour, bufFrom + 5);

        // envoyer un mail  l'expditeur, sauf si option -s
        if (! opts && (bufFrom [0] || bufRetPath [0]))
            if (avertir ())
                avertis ++;
            else
                non_avertis ++;
    }
    else
        conserves ++;
}


/* avertit l'expditeur d'un mail en HTML pur que son message est refus */

int avertir ()
{
    char bufw [120];  // buffer d'envoi d'une ligne ou d'une requte smtp


    // l'envoi du mail d'avertissement suppose l'accs  un fichier de donnes
    if (! acces_mailrep ("suphtm"))
        return (0);

    // expditeur du message
    sprintf (bufw, "MAIL FROM: <%s>", adr_mail);
    envoie_smtp (bufw);
    lire_smtp ();

    // destinataire du message
    strcpy (bufw, "RCPT TO: <");

    // on envoie en priorit  l'adresse du Return_Path
    if (bufRetPath [0])
        ajout_adr (bufw, bufRetPath + 12);
    // sinon  l'expditeur dclar
    else
        ajout_adr (bufw, bufFrom + 5);

    strcat (bufw, ">");
    envoie_smtp (bufw);
    lire_smtp ();

    // si destinataire refus, on le signale
    if (buf_lect [0] != '2')
    {
        // Affichage de l'adresse  problme
        // "Destinataire %s refus\n"
        putchar ('\n');
        printf (message ("REFUS_DEST"), bufw + 9);
        puts (buf_lect);

        // on n'envoie pas de mail
        envoie_smtp ("RSET");
        lire_smtp ();
        return 0;  // mail non envoy
    }

    // prparation de l'envoi du message
    envoie_smtp ("DATA");
    lire_smtp ();

    // gnration de l'entte du message

    // si prfixe, l'adresse expditeur est modifie dans
    // l'entte du message pour tromper les robots de spam
    if (*pref_adr)
        sprintf (bufw, "From: <%s-%s>", pref_adr, adr_mail);
    else
        sprintf (bufw, "From: <%s>", adr_mail);

    envoie_smtp (bufw);
    sprintf (bufw, "To%s", bufFrom + 4);
    envoie_smtp (bufw);

    if (*bufReply && strcmp (bufFrom + 5, bufReply + 9) != 0)
    {
        sprintf (bufw, "Cc%s", bufReply + 8);
        envoie_smtp (bufw);
    }

    // "Subject: Votre message du ..."
    sprintf (bufw, "Subject: %s %s", chaine_mailrep ("SUJET"), bufDate + 4);
    envoie_smtp (bufw);

    // "Content-Type: text/plain; charset=..."
    sprintf (bufw, "Content-Type: text/plain; charset=%s",
                                  chaine_mailrep ("CHARSET"));
    envoie_smtp (bufw);

    envoie_smtp ("Content-Transfer-Encoding: 8bit");

    // "filtre suphtm : logiciel libre multilingue"
    sprintf (bufw, "User-Agent: filtre %s : %s\n", nomcom (),
                                message ("QUALIF_LIBREMAIL"));
    envoie_smtp (bufw);
    envoie_smtp ("");

    // "Vous m'avez envoy un mail ayant les caractristiques suivantes :"
    envoie_smtp (chaine_mailrep ("ANNONCE"));
    envoie_smtp ("");

    // rappel du message concern
    transf_ligne (bufDate);
    transf_ligne (bufFrom);
    transf_ligne (bufTo);
    transf_ligne (bufCc);
    transf_ligne (bufReply);
    transf_ligne (bufSubject);
    transf_ligne (bufContent);
    envoie_smtp ("");

    // envoi du message explicatif
    envoi_mailrep ();

    // indicateur de fin du message
    envoie_smtp (".");
    lire_smtp ();

    return 1;   // bon droulement
}


/* rcupre l'adresse mail entre les < > s'ils
   existent et la rajoute au buffer destination */

void ajout_adr (char *dest, char *source)
{
    int orig, debut, posdest;

    orig = 0;

    // chercher le @
    while (source [orig] != '@' && source [orig])
        orig ++;

    // remonter au debut de l'adresse
    while (orig > 0 && source [orig] != ' ' && source [orig] != '<')
        orig--;

    // l'adresse commence sur un caractre significatif
    if (source [orig] == '<' || source [orig] == ' ')
        debut = orig + 1;
    else
        debut = orig;

    // copie de l'adresse
    posdest = strlen (dest);

    // on s'arrte si l'on rencontre un  > , un blanc ou une fin de chaine
    while (source [debut] != '>' && source [debut] & 0xDF)
        dest [posdest++] = source [debut++];

    // terminer l'expression de l'adresse
    dest [posdest] = '\0';
}


/* copie dans le mail  envoyer d'une ligne du mail reu */

void transf_ligne (char *message)
{
    char bufw [82];  // buffer contenant la ligne  envoyer

    if (*message)
    {
        strcpy (bufw, "> ");
        strcpy (bufw + 2, message);
        envoie_smtp (bufw);
    }
}
